import pandas as pd
import numpy as np
from collections import OrderedDict
from patsy import dmatrices, dmatrix
import re
import warnings
import statsmodels.api as sm
import matplotlib.pyplot as plt
from copy import deepcopy
from bambi.external.six import string_types
from bambi.external.patsy import Ignore_NA, rename_columns
from bambi.priors import PriorFactory, PriorScaler, Prior
from bambi.utils import listify
import pymc3 as pm


class Model(object):

    '''
    Args:
        data (DataFrame, str): the dataset to use. Either a pandas
            DataFrame, or the name of the file containing the data, which
            will be passed to pd.read_table().
        default_priors (dict, str): An optional specification of the
            default priors to use for all model terms. Either a dict
            containing named distributions, families, and terms (see the
            documentation in priors.PriorFactory for details), or the name
            of a JSON file containing the same information.
        auto_scale (bool): If True (default), priors are automatically rescaled
            to the data (to be weakly informative) any time default priors are
            used. Note that any priors explicitly set by the user will always
            take precedence over default priors.
        dropna (bool): When True, rows with any missing values in either the
            predictors or outcome are automatically dropped from the dataset in
            a listwise manner.
        taylor (int): Order of Taylor expansion to use in approximate variance
            when constructing the default priors. Should be between 1 and 13.
            Lower values are less accurate, tending to undershoot the correct
            prior width, but are faster to compute and more stable. Odd-
            numbered values tend to work better. Defaults to 5 for Normal
            models and 1 for non-Normal models. Values higher than the defaults
            are generally not recommended as they can be unstable.
        noncentered (True): If True (default), uses a non-centered
            parameterization for normal hyperpriors on grouped parameters.
            If False, naive (centered) parameterization is used.
    '''

    def __init__(self, data=None, default_priors=None, auto_scale=True,
                 dropna=False, taylor=None, noncentered=True):

        if isinstance(data, string_types):
            data = pd.read_table(data, sep=None)

        self.default_priors = PriorFactory(default_priors)

        obj_cols = data.select_dtypes(['object']).columns
        data[obj_cols] = data[obj_cols].apply(lambda x: x.astype('category'))
        self.data = data
        # Some random effects stuff later requires us to make guesses about
        # column groupings into terms based on patsy's naming scheme.
        if re.search("[\[\]]+", ''.join(data.columns)):
            warnings.warn("At least one of the column names in the specified "
                          "dataset contain square brackets ('[' or ']')."
                          "This may cause unexpected behavior if you specify "
                          "models with random effects. You are encouraged to "
                          "rename your columns to avoid square brackets.")
        self.reset()

        self.auto_scale = auto_scale
        self.dropna = dropna
        self.taylor = taylor
        self.noncentered = noncentered
        self._backend_name = None

    def reset(self):
        '''
        Reset list of terms and y-variable.
        '''
        self.terms = OrderedDict()
        self.y = None
        self.backend = None

    def _set_backend(self, backend):

        backend = backend.lower()

        if backend.startswith('pymc'):
            from bambi.backends import PyMC3BackEnd
            self.backend = PyMC3BackEnd()
        elif backend == 'stan':
            from bambi.backends import StanBackEnd
            self.backend = StanBackEnd()
        else:
            raise ValueError(
                "At the moment, only the PyMC3 and Stan backends are "
                "supported.")

        self._backend_name = backend

    def build(self, backend=None):
        ''' Set up the model for sampling/fitting. Performs any steps that
        require access to all model terms (e.g., scaling priors on each term),
        then calls the BackEnd's build() method.
        Args:
            backend (str): The name of the backend to use for model fitting.
                Currently, 'pymc' and 'stan' are supported. If None, assume
                that fit() has already been called (possibly without building),
                and look in self._backend_name.
        '''

        if backend is None:
            if self._backend_name is None:
                raise ValueError("Error: no backend was passed or set in the "
                                 "Model; did you forget to call fit()?")
            backend = self._backend_name

        if self.y is None:
            raise ValueError("No outcome (y) variable is set! Please specify "
                             "an outcome variable using the formula interface "
                             "before build() or fit().")

        # Check for NaNs and halt if dropna is False--otherwise issue warning.
        arrs = []
        for t in self.terms.values():
            arrs.append(t.data)
        X = np.concatenate(arrs + [self.y.data], axis=1)
        na_index = np.isnan(X).any(1)
        if na_index.sum():
            msg = "%d rows were found contain at least one missing value." \
                % na_index.sum()
            if not self.dropna:
                msg += "Please make sure the dataset contains no missing " \
                       "values. Alternatively, if you want rows with missing "\
                       "values to be automatically deleted in a list-wise " \
                       "manner (not recommended), please set dropna=True at " \
                       "model initialization."
                raise ValueError(msg)

            # warn and then remove missing values
            msg += " Automatically removing %d rows from the dataset." \
                % na_index.sum()
            warnings.warn(msg)
            keeps = np.invert(na_index)
            # removing missing rows
            for t in self.terms.values():
                t.data = t.data[keeps]
                # alter additional attributes in RandomTerms
                if isinstance(t, RandomTerm):
                    t.grouper = t.grouper[keeps]
                    t.predictor = t.predictor[keeps]
                    t.group_index = t._invert_dummies(t.grouper)
            self.y.data = self.y.data[keeps]

        # X = fixed effects design matrix (excluding intercept/constant term)
        # r2_x = 1 - 1/VIF, i.e., R2 for predicting each x from all other x's.
        # only compute these stats if there are multiple terms in the model
        terms = [t for t in self.fixed_terms.values() if t.name != 'Intercept']

        if len(self.fixed_terms) > 1:

            X = [pd.DataFrame(x.data, columns=x.levels) for x in terms]
            X = pd.concat(X, axis=1)

            self.dm_statistics = {
                'r2_x': pd.Series({
                    x: sm.OLS(endog=X[x],
                              exog=sm.add_constant(X.drop(x, axis=1))
                              if 'Intercept' in self.term_names
                              else X.drop(x, axis=1)).fit().rsquared
                    for x in list(X.columns)}),
                'sd_x': X.std(),
                'mean_x': X.mean(axis=0)
            }

            # save potentially useful info for diagnostics, send to
            # ModelResults.
            # mat = correlation matrix of X, w/ diagonal replaced by X means
            mat = X.corr()
            for x in list(mat.columns):
                mat.loc[x, x] = self.dm_statistics['mean_x'][x]
            self._diagnostics = {
                # the Variance Inflation Factors (VIF), which is possibly
                # useful for diagnostics
                'VIF': 1/(1 - self.dm_statistics['r2_x']),
                'corr_mean_X': mat
            }

            # throw informative error if perfect collinearity among fixed fx
            if any(self.dm_statistics['r2_x'] > .999):
                raise ValueError(
                    "There is perfect collinearity among the fixed effects!\n"
                    "Printing some design matrix statistics:\n" +
                    str(self.dm_statistics) + '\n' +
                    str(self._diagnostics))

        # throw informative error message if any categorical predictors have 1
        # category
        num_cats = [x.data.size for x in self.fixed_terms.values()]
        if any(np.array(num_cats) == 0):
            raise ValueError(
                "At least one categorical predictor contains only 1 category!")

        # only set priors if there is at least one term in the model
        if len(self.terms) > 0:
            # Get and scale default priors if none are defined yet
            if self.taylor is not None:
                taylor = self.taylor
            else:
                taylor = 5 if self.family.name == 'gaussian' else 1
            scaler = PriorScaler(self, taylor=taylor)
            scaler.scale()

        # For bernoulli models with n_trials = 1 (most common use case),
        # tell user which event is being modeled
        if self.family.name == 'bernoulli' and np.max(self.y.data) < 1.01:
            event = next(
                i for i, x in enumerate(self.y.data.flatten()) if x > .99)
            warnings.warn('Modeling the probability that {}==\'{}\''.format(
                self.y.name, str(self.data[self.y.name].iloc[event])))

        self._set_backend(backend)
        self.backend.build(self)
        self.built = True

    def fit(self, fixed=None, random=None, priors=None, family='gaussian',
            link=None, run=True, categorical=None, backend=None, **kwargs):
        '''
        Fit the model using the specified BackEnd.
        Args:
            fixed (str): Optional formula specification of fixed effects.
            random (list): Optional list-based specification of random effects.
            priors (dict): Optional specification of priors for one or more
                terms. A dict where the keys are the names of terms in the
                model, and the values are either instances of class Prior or
                ints, floats, or strings that specify the width of the priors
                on a standardized scale.
            family (str, Family): A specification of the model family
                (analogous to the family object in R). Either a string, or an
                instance of class priors.Family. If a string is passed, a
                family with the corresponding name must be defined in the
                defaults loaded at Model initialization. Valid pre-defined
                families are 'gaussian', 'bernoulli', 'poisson', and 't'.
            link (str): The model link function to use. Can be either a string
                (must be one of the options defined in the current backend;
                typically this will include at least 'identity', 'logit',
                'inverse', and 'log'), or a callable that takes a 1D ndarray
                or theano tensor as the sole argument and returns one with
                the same shape.
            run (bool): Whether or not to immediately begin fitting the model
                once any set up of passed arguments is complete.
            categorical (str, list): The names of any variables to treat as
                categorical. Can be either a single variable name, or a list
                of names. If categorical is None, the data type of the columns
                in the DataFrame will be used to infer handling. In cases where
                numeric columns are to be treated as categoricals (e.g., random
                factors coded as numerical IDs), explicitly passing variable
                names via this argument is recommended.
            backend (str): The name of the BackEnd to use. Currently only
                'pymc' and 'stan' backends are supported. Defaults to PyMC3.
        '''
        if fixed is not None or random is not None:
            self.add(fixed=fixed, random=random, priors=priors, family=family,
                     link=link, categorical=categorical, append=False)

        ''' Run the BackEnd to fit the model. '''
        if backend is None:
            backend = 'pymc' if self._backend_name is None else self._backend_name

        if run:
            if not self.built or backend != self._backend_name:
                warnings.warn("Current Bayesian model has not been built yet "
                              "with the %s back-end; building it first before "
                              "sampling begins." % self._backend_name)
                self.build(backend)
            return self.backend.run(**kwargs)

        self._backend_name = backend

    def add(self, fixed=None, random=None, priors=None, family='gaussian',
            link=None, categorical=None, append=True):
        '''
        Adds one or more terms to the model via an R-like formula syntax.
        Args:
            fixed (str): Optional formula specification of fixed effects.
            random (list): Optional list-based specification of random effects.
            priors (dict): Optional specification of priors for one or more
                terms. A dict where the keys are the names of terms in the
                model, and the values are either instances of class Prior or
                ints, floats, or strings that specify the width of the priors
                on a standardized scale.
            family (str, Family): A specification of the model family
                (analogous to the family object in R). Either a string, or an
                instance of class priors.Family. If a string is passed, a
                family with the corresponding name must be defined in the
                defaults loaded at Model initialization. Valid pre-defined
                families are 'gaussian', 'bernoulli', 'poisson', and 't'.
            link (str): The model link function to use. Can be either a string
                (must be one of the options defined in the current backend;
                typically this will include at least 'identity', 'logit',
                'inverse', and 'log'), or a callable that takes a 1D ndarray
                or theano tensor as the sole argument and returns one with
                the same shape.
            categorical (str, list): The names of any variables to treat as
                categorical. Can be either a single variable name, or a list
                of names. If categorical is None, the data type of the columns
                in the DataFrame will be used to infer handling. In cases where
                numeric columns are to be treated as categoricals (e.g., random
                factors coded as numerical IDs), explicitly passing variable
                names via this argument is recommended.
            append (bool): if True, terms are appended to the existing model
                rather than replacing any existing terms. This allows
                formula-based specification of the model in stages.
        '''
        data = self.data

        # Primitive values (floats, strs) can be overwritten with Prior objects
        # so we need to make sure to copy first to avoid bad things happening
        # if user is re-using same prior dict in multiple models.
        if priors is None:
            priors = {}
        else:
            priors = deepcopy(priors)

        if not append:
            self.reset()

        # Explicitly convert columns to category if desired--though this
        # can also be done within the formula using C().
        if categorical is not None:
            data = data.copy()
            cats = listify(categorical)
            data[cats] = data[cats].apply(lambda x: x.astype('category'))

        if fixed is not None:
            if '~' in fixed:
                # check to see if formula is using the 'y[event] ~ x' syntax
                # (for bernoulli models). If so, chop it into groups:
                # 1 = 'y[event]', 2 = 'y', 3 = 'event', 4 = 'x'
                # If this syntax is not being used, event = None
                event = re.match(r'^((\S+)\[(\S+)\])\s*~(.*)$', fixed)
                if event is not None:
                    fixed = '{}~{}'.format(event.group(2), event.group(4))
                y, X = dmatrices(fixed, data=data, NA_action=Ignore_NA())
                y_label = y.design_info.term_names[0]
                if event is not None:
                    # pass in new Y data that has 1 if y=event and 0 otherwise
                    y_data = y[:, y.design_info.column_names.index(event.group(1))]
                    y_data = pd.DataFrame({event.group(3): y_data})
                    self._add_y(y_label, family=family, link=link, data=y_data)
                else:
                    # use Y as-is
                    self._add_y(y_label, family=family, link=link)
            else:
                X = dmatrix(fixed, data=data, NA_action=Ignore_NA())

            # Loop over predictor terms
            for _name, _slice in X.design_info.term_name_slices.items():
                cols = X.design_info.column_names[_slice]
                term_data = pd.DataFrame(X[:, _slice], columns=cols)
                prior = priors.pop(_name, priors.get('fixed', None))
                self.terms[_name] = Term(self, _name, term_data, prior=prior)

        # Random effects
        if random is not None:

            random = listify(random)

            for f in random:

                f = f.strip()

                # Split specification into intercept, predictor, and grouper
                patt = r'^([01]+)*[\s\+]*([^\|]+)*\|(.*)'

                intcpt, pred, grpr = re.search(patt, f).groups()
                label = '{}|{}'.format(pred, grpr) if pred else grpr
                prior = priors.pop(label, priors.get('random', None))

                # Treat all grouping variables as categoricals, regardless of
                # their dtype and what the user may have specified in the
                # 'categorical' argument.
                var_names = re.findall('(\w+)', grpr)
                for v in var_names:
                    if v in data.columns:
                        data[v] = data.loc[:, v].astype('category')
                        self.data[v] = data[v]

                # Default to including random intercepts
                intcpt = 1 if intcpt is None else int(intcpt)

                grpr_df = dmatrix('0+%s' % grpr, data, return_type='dataframe',
                                  NA_action=Ignore_NA())

                # If there's no predictor, we must be adding random intercepts
                if not pred and grpr not in self.terms:
                    name = '1|' + grpr
                    pred = np.ones((len(grpr_df), 1))
                    term = RandomTerm(self, name, grpr_df, pred, grpr_df.values,
                                      categorical=True, prior=prior)
                    self.terms[name] = term
                else:
                    pred_df = dmatrix('%s+%s' % (intcpt, pred), data,
                                      return_type='dataframe',
                                      NA_action=Ignore_NA())
                    # determine value of the 'constant' attribute
                    const = np.atleast_2d(pred_df.T).T.sum(1).var() == 0

                    for col, i in pred_df.design_info.column_name_indexes.items():
                        pred_data = pred_df.iloc[:, i]
                        lev_data = grpr_df.multiply(pred_data, axis=0)

                        # Also rename intercepts and skip if already added.
                        # This can happen if user specifies something like
                        # random=['1|school', 'student|school'].
                        if col == 'Intercept':
                            if grpr in self.terms:
                                continue
                            label = '1|%s' % grpr
                        else:
                            label = col + '|' + grpr

                        prior = priors.pop(label, priors.get('random', None))

                        # Categorical or continuous is determined from data
                        ld = lev_data.values
                        if ((ld == 0) | (ld == 1)).all():
                            lev_data = lev_data.astype(int)
                            cat = True
                        else:
                            cat = False

                        pred_data = pred_data[:, None]  # Must be 2D later
                        term = RandomTerm(self, label, lev_data, pred_data,
                                          grpr_df.values, categorical=cat,
                                          constant=const if const else None)
                        self.terms[label] = term

        self.built = False

    def _add_y(self, variable, prior=None, family='gaussian', link=None, *args,
               **kwargs):
        '''
        Add a dependent (or outcome) variable to the model.
        Args:
            variable (str): the name of the dataset column containing the
                y values.
            prior (Prior, int, float, str): Optional specification of prior.
                Can be an instance of class Prior, a numeric value, or a string
                describing the width. In the numeric case, the distribution
                specified in the defaults will be used, and the passed value
                will be used to scale the appropriate variance parameter. For
                strings (e.g., 'wide', 'narrow', 'medium', or 'superwide'),
                predefined values will be used.
            family (str, Family): A specification of the model family
                (analogous to the family object in R). Either a string, or an
                instance of class priors.Family. If a string is passed, a
                family with the corresponding name must be defined in the
                defaults loaded at Model initialization. Valid pre-defined
                families are 'gaussian', 'bernoulli', 'poisson', and 't'.
            link (str): The model link function to use. Can be either a string
                (must be one of the options defined in the current backend;
                typically this will include at least 'identity', 'logit',
                'inverse', and 'log'), or a callable that takes a 1D ndarray
                or theano tensor as the sole argument and returns one with
                the same shape.
            args, kwargs: Optional positional and keyword arguments to pass
                onto Term initializer.
        '''
        if isinstance(family, string_types):
            family = self.default_priors.get(family=family)
        self.family = family

        # Override family's link if another is explicitly passed
        if link is not None:
            self.family.link = link

        if prior is None:
            prior = self.family.prior

        # implement default Uniform [0, sd(Y)] prior for residual SD
        if self.family.name == 'gaussian':
            prior.update(sd=Prior('Uniform', lower=0,
                                  upper=self.data[variable].std()))

        data = kwargs.pop('data', self.data[variable])
        term = Term(self, variable, data, prior=prior, *args, **kwargs)
        self.y = term
        self.built = False

    def _match_derived_terms(self, name):
        ''' Returns all (random) terms whose named are derived from the
        specified string. For example, 'condition|subject' should match the
        terms with names '1|subject', 'condition[T.1]|subject', and so on.
        Only works for strings with grouping operator ('|').
        '''
        if '|' not in name:
            return None

        patt = r'^([01]+)*[\s\+]*([^\|]+)*\|(.*)'
        intcpt, pred, grpr = re.search(patt, name).groups()

        intcpt = '1|%s' % grpr
        if not pred:
            return [self.terms[intcpt]] if intcpt in self.terms else None

        source = '%s|%s' % (pred, grpr)
        found = [t for (n, t) in self.terms.items() if n == intcpt or
                 re.sub('(\[.*?\])', '', n) == source]
        # If only the intercept matches, return None, because we want to err
        # on the side of caution and not consider '1|subject' to be a match for
        # 'condition|subject' if no slopes are found (e.g., the intercept could
        # have been set by some other specification like 'gender|subject').
        return found if found and (len(found) > 1 or found[0].name != intcpt) \
            else None

    def set_priors(self, priors=None, fixed=None, random=None,
                   match_derived_names=True):
        '''
        Set priors for one or more existing terms.
        Args:
            priors (dict): Dict of priors to update. Keys are names of terms
                to update; values are the new priors (either a Prior instance,
                or an int or float that scales the default priors). Note that
                a tuple can be passed as the key, in which case the same prior
                will be applied to all terms named in the tuple.
            fixed (Prior, int, float, str): a prior specification to apply to
                all fixed terms currently included in the model.
            random (Prior, int, float, str): a prior specification to apply to
                all random terms currently included in the model.
            match_derived_names (bool): if True, the specified prior(s) will be
                applied not only to terms that match the keyword exactly,
                but to the levels of random effects that were derived from
                the original specification with the passed name. For example,
                `priors={'condition|subject':0.5}` would apply the prior
                to the terms with names '1|subject', 'condition[T.1]|subject',
                and so on. If False, an exact match is required for the
                prior to be applied.
        '''

        targets = {}

        if fixed is not None:
            targets.update({name: fixed for name in self.fixed_terms.keys()})

        if random is not None:
            targets.update({name: random for name in self.random_terms.keys()})

        if priors is not None:
            for k, prior in priors.items():
                for name in listify(k):
                    term_names = list(self.terms.keys())
                    msg = "No terms in model match '%s'." % name
                    if name not in term_names:
                        terms = self._match_derived_terms(name)
                        if not match_derived_names or terms is None:
                            raise ValueError(msg)
                        for t in terms:
                            targets[t.name] = prior
                    else:
                        targets[name] = prior

        for prior in targets.values():
            if isinstance(prior, Prior):
                prior._auto_scale = False

        for name, prior in targets.items():
            self.terms[name].prior = prior

        if fixed is not None or random is not None or priors is not None:
            self.built = False

    def plot(self, varnames=None):
        self.plot_priors(varnames)

    def plot_priors(self, varnames=None):
        if not self.built:
            raise ValueError("Cannot plot priors until model is built!")

        with pm.Model():
            # get priors for fixed fx, separately for each level of each
            # predictor
            dists = []
            for t in self.fixed_terms.values():
                if varnames is not None and t.name not in varnames:
                    continue
                for i, l in enumerate(t.levels):
                    params = {k: v[i % len(v)]
                              if isinstance(v, np.ndarray) else v
                              for k, v in t.prior.args.items()}
                    dists += [getattr(pm, t.prior.name)(l, **params)]

            # get priors for random effect SDs
            for t in self.random_terms.values():
                if varnames is not None and t.name not in varnames:
                    continue
                prior = t.prior.args['sd'].name
                params = t.prior.args['sd'].args
                dists += [getattr(pm, prior)(t.name+'_sd', **params)]

            # add priors on Y params if applicable
            y_prior = [(k, v) for k, v in self.y.prior.args.items()
                       if isinstance(v, Prior)]
            if len(y_prior):
                for p in y_prior:
                    dists += [getattr(pm, p[1].name)('_'.join([self.y.name,
                                                               p[0]]), **p[1].args)]

            # make the plot!
            p = float(len(dists))
            fig, axes = plt.subplots(int(np.ceil(p/2)), 2,
                                     figsize=(12, np.ceil(p/2)*2))
            # in case there is only 1 row
            if int(np.ceil(p/2)) < 2:
                axes = axes[None, :]
            for i, d in enumerate(dists):
                dist = d.distribution if isinstance(d, pm.model.FreeRV) else d
                samp = pd.Series(dist.random(size=1000).flatten())
                samp.plot(kind='hist', ax=axes[divmod(i, 2)[0], divmod(i, 2)[1]],
                          normed=True)
                samp.plot(kind='kde', ax=axes[divmod(i, 2)[0], divmod(i, 2)[1]],
                          color='b')
                axes[divmod(i, 2)[0], divmod(i, 2)[1]].set_title(d.name)
            fig.tight_layout()

        return axes

    @property
    def term_names(self):
        ''' Return names of all terms in order of addition to model. '''
        return list(self.terms.keys())

    @property
    def fixed_terms(self):
        ''' Return dict of all and only fixed effects in model. '''
        return {k: v for (k, v) in self.terms.items() if not v.random}

    @property
    def random_terms(self):
        ''' Return dict of all and only random effects in model. '''
        return {k: v for (k, v) in self.terms.items() if v.random}


class Term(object):

    '''
    Representation of a single (fixed) model term.
    Args:
        name (str): Name of the term.
        data (DataFrame, Series, ndarray): The term values.
        categorical (bool): If True, the source variable is interpreted as
            nominal/categorical. If False, the source variable is treated
            as continuous.
        prior (Prior): A specification of the prior(s) to use. An instance
            of class priors.Prior.
        constant (bool): indicates whether the term levels collectively
            act as a constant, in which case the term is treated as an
            intercept for prior distribution purposes.
    '''
    random = False

    def __init__(self, model, name, data, categorical=False, prior=None,
                 constant=None):

        self.model = model
        self.name = name
        self.categorical = categorical
        self._reduced_data = None

        if isinstance(data, pd.Series):
            data = data.to_frame()
        if isinstance(data, pd.DataFrame):
            self.levels = list(data.columns)
            data = data.values

        # Random effects pass through here
        else:
            data = np.atleast_2d(data)
            self.levels = list(range(data.shape[1]))

        self.data = data

        # identify and flag intercept and cell-means terms (i.e., full-rank
        # dummy codes), which receive special priors
        if constant is None:
            self.constant = np.atleast_2d(data.T).T.sum(1).var() == 0
        else:
            self.constant = constant

        self.set_prior(prior)

    def set_prior(self, prior):
        _type = 'intercept' if self.name == 'Intercept' else \
                'random' if self.random else 'fixed'

        if prior is None and not self.model.auto_scale:
            prior = self.model.default_priors.get(term=_type + '_flat')

        if isinstance(prior, Prior):
            prior._auto_scale = False
        else:
            _scale = prior
            prior = self.model.default_priors.get(term=_type)
            prior.scale = _scale

        self.prior = prior


class RandomTerm(Term):

    random = True

    def __init__(self, model, name, data, predictor, grouper,
                 categorical=False, prior=None, constant=None):

        super(RandomTerm, self).__init__(model, name, data, categorical, prior,
              constant)
        self.grouper = grouper
        self.predictor = predictor
        self.group_index = self._invert_dummies(grouper)

    def _invert_dummies(self, dummies):
        ''' For the sake of computational efficiency (i.e., to avoid lots of
        large matrix multiplications in the backends), invert the dummy-coding
        process and represent full-rank dummies as a vector of indices into the
        coefficients. '''
        vec = np.zeros(len(dummies), dtype=int)
        for i in range(1, dummies.shape[1]):
            vec[dummies[:, i] == 1] = i
        return vec



from __future__ import absolute_import
import numpy as np
import re
from patsy.util import safe_scalar_isnan


class Ignore_NA(object):
    """
    Custom patsy.missing.NAAction class to force Patsy to ignore missing
    values. See Patsy code/API for NAAction documentation.
    """
    def __init__(self, on_NA="ignore", NA_types=["None", "NaN"]):
        self.on_NA = on_NA
        if isinstance(NA_types, str):
            raise ValueError("NA_types should be a list of strings")
        self.NA_types = tuple(NA_types)

    def is_categorical_NA(self, obj):
        if "NaN" in self.NA_types and safe_scalar_isnan(obj):
            return True
        if "None" in self.NA_types and obj is None:
            return True
        return False

    def is_numerical_NA(self, arr):
        mask = np.zeros(arr.shape, dtype=bool)
        if "NaN" in self.NA_types:
            mask |= np.isnan(arr)
        if mask.ndim > 1:
            mask = np.any(mask, axis=1)
        return mask

    def handle_NA(self, values, is_NAs, origins):
        return values


def rename_columns(columns, name_lists):
    ''' Renames numerical indices in column names returned by patsy dmatrix /
    dmatrices calls based on the corresponding string levels.
    Args:
        columns (list): List of cols from dmatrix's .design_info.column_names
        name_lists (list): List of lists, where the i'th list gives the set of
            level names for the i'th index that needs to be replaced.
    Example:
        ts = ['threecats[1]:subjects[4]', 'threecats[0]:subjects[3]']
        rename_columns(ts, (['a', 'b', 'c'], ['d1', 'd2', 'd3', 'd4', 'd5']))
        # Returns ['threecats[b]:subjects[d5]', 'threecats[a]:subjects[d4]']
    '''
    def _replace(c, args):
        grps = re.findall('([^\]]*)(\[(\d+)\])', c)
        for i, (prefix, box, ind) in enumerate(grps):
            c = c.replace(prefix + box, prefix + '[%s]' % args[i][int(ind)])
        return c
    return [_replace(c, name_lists) for c in columns]



from .models import Model
from .priors import Prior, Family
from .results import ModelResults
from .backends import StanBackEnd, PyMC3BackEnd


__all__ = [
    'Model',
    'Prior',
    'Family',
    'ModelResults',
    'StanBackEnd',
    'PyMC3BackEnd',
]



import numpy as np
import pandas as pd

# code adapted from pymc3.diagnostics:
# https://github.com/pymc-devs/pymc3/blob/master/pymc3/diagnostics.py


def _vhat_w(mcmc):
    # Calculate between-chain variance
    B = mcmc.n_samples * mcmc.data.mean(axis=0).var(axis=0, ddof=1)

    # Calculate within-chain variance
    W = mcmc.data.var(axis=0, ddof=1).mean(axis=0)

    # Estimate of marginal posterior variance
    Vhat = W * (mcmc.n_samples - 1) / mcmc.n_samples + B / mcmc.n_samples

    return Vhat, W


def gelman_rubin(mcmc):
    """
    mcmc (MCMCResults): Pre-sliced MCMC samples to compute diagnostics for.
    """

    if mcmc.n_chains < 2:
        raise ValueError(
            'Gelman-Rubin diagnostic requires multiple chains '
            'of the same length.')

    # get Vhat and within-chain variance
    Vhat, W = _vhat_w(mcmc)

    # compute and return Gelman-Rubin statistic
    Rhat = np.sqrt(Vhat / W)
    return pd.DataFrame({'gelman_rubin': Rhat}, index=mcmc.levels)


def effective_n(mcmc):
    """
    mcmc (MCMCResults): Pre-sliced MCMC samples to compute diagnostics for.
    """

    if mcmc.n_chains < 2:
        raise ValueError(
            'Calculation of effective sample size requires multiple chains '
            'of the same length.')

    def get_neff(x, Vhat):
        n_samples = x.shape[0]
        n_chains = x.shape[1]

        negative_autocorr = False
        t = 1

        rho = np.ones(n_samples)
        # Iterate until the sum of consecutive estimates of autocorrelation is
        # negative
        while not negative_autocorr and (t < n_samples):
            variogram = np.mean((x[t:, :] - x[:-t, :])**2)
            rho[t] = 1. - variogram / (2. * Vhat)
            negative_autocorr = sum(rho[t - 1:t + 1]) < 0
            t += 1

        if t % 2:
            t -= 1

        return min(n_chains * n_samples,
                   int(n_chains * n_samples / (1. + 2 * rho[1:t-1].sum())))

    Vhat, W = _vhat_w(mcmc)
    n_eff = [get_neff(mcmc.data[:, :, i], x) for i, x in enumerate(Vhat)]

    return pd.DataFrame({'effective_n': n_eff}, index=mcmc.levels)



import numpy as np
import pandas as pd
from scipy.special import hyp2f1
from os.path import dirname, join
from copy import deepcopy
import json
import re
import statsmodels.api as sm
from bambi.external.six import string_types


class Family(object):

    '''
    A specification of model family.
    Args:
        name (str): Family name
        prior (Prior): A Prior instance specifying the model likelihood prior
        link (str): The name of the link function transforming the linear
            model prediction to a parameter of the likelihood
        parent (str): The name of the prior parameter to set to the link-
            transformed predicted outcome (e.g., mu, p, etc.).
    '''

    def __init__(self, name, prior, link, parent):
        self.name = name
        self.prior = prior
        self.link = link
        self.parent = parent
        fams = {
            'gaussian': sm.families.Gaussian,
            'bernoulli': sm.families.Binomial,
            'poisson': sm.families.Poisson,
            't': None  # not implemented in statsmodels
        }
        self.smfamily = fams[name] if name in fams.keys() else None


class Prior(object):

    '''
    Abstract specification of a term prior.
    Args:
        name (str): Name of prior distribution (e.g., Normal, Bernoulli, etc.)
        kwargs (dict): Optional keywords specifying the parameters of the
            named distribution.
    '''

    def __init__(self, name, scale=None, **kwargs):
        self.name = name
        self._auto_scale = True
        self.scale = scale
        self.args = {}
        self.update(**kwargs)

    def update(self, **kwargs):
        '''
        Update the model arguments with additional arguments.
        Args:
            kwargs (dict): Optional keyword arguments to add to prior args.
        '''
        self.args.update(kwargs)


class PriorFactory(object):

    '''
    An object that supports specification and easy retrieval of default priors.
    Args:
        defaults (str, dict): Optional base configuration containing default
            priors for distribution, families, and term types. If a string,
            the name of a JSON file containing the config. If a dict, must
            contain keys for 'dists', 'terms', and 'families'; see the built-in
            JSON configuration for an example. If None, a built-in set of
            priors will be used as defaults.
        dists (dict): Optional specification of named distributions to use
            as priors. Each key gives the name of a newly defined distribution;
            values are two-element lists, where the first element is the name
            of the built-in distribution to use ('Normal', 'Cauchy', etc.),
            and the second element is a dictionary of parameters on that
            distribution (e.g., {'mu': 0, 'sd': 10}). Priors can be nested
            to arbitrary depths by replacing any parameter with another prior
            specification.
        terms (dict): Optional specification of default priors for different
            model term types. Valid keys are 'intercept', 'fixed', or 'random'.
            Values are either strings preprended by a #, in which case they
            are interpreted as pointers to distributions named in the dists
            dictionary, or key -> value specifications in the same format as
            elements in the dists dictionary.
        families (dict): Optional specification of default priors for named
            family objects. Keys are family names, and values are dicts
            containing mandatory keys for 'dist', 'link', and 'parent'.

    Examples:
        >>> dists = { 'my_dist': ['Normal', {'mu': 10, 'sd': 1000}]}
        >>> pf = PriorFactory(dists=dists)

        >>> families = { 'normalish': { 'dist': ['normal', {sd: '#my_dist'}],
        >>>                             link:'identity', parent: 'mu'}}
        >>> pf = PriorFactory(dists=dists, families=families)
    '''

    def __init__(self, defaults=None, dists=None, terms=None, families=None):

        if defaults is None:
            defaults = join(dirname(__file__), 'config', 'priors.json')

        if isinstance(defaults, string_types):
            defaults = json.load(open(defaults, 'r'))

        # Just in case the user plans to use the same defaults elsewhere
        defaults = deepcopy(defaults)

        if isinstance(dists, dict):
            defaults['dists'].update(dists)

        if isinstance(terms, dict):
            defaults['terms'].update(terms)

        if isinstance(families, dict):
            defaults['families'].update(families)

        self.dists = defaults['dists']
        self.terms = defaults['terms']
        self.families = defaults['families']

    def _get_prior(self, spec, **kwargs):

        if isinstance(spec, string_types):
            spec = re.sub('^\#', '', spec)
            return self._get_prior(self.dists[spec])
        elif isinstance(spec, (list, tuple)):
            name, args = spec
            if name.startswith('#'):
                name = re.sub('^\#', '', name)
                prior = self._get_prior(self.dists[name])
            else:
                prior = Prior(name, **kwargs)
            args = {k: self._get_prior(v) for (k, v) in args.items()}
            prior.update(**args)
            return prior
        else:
            return spec

    def get(self, dist=None, term=None, family=None):
        '''
        Retrieve default prior for a named distribution, term type, or family.
        Args:
            dist (str): Name of desired distribution. Note that the name is
                the key in the defaults dictionary, not the name of the
                Distribution object used to construct the prior.
            term (str): The type of term family to retrieve defaults for.
                Must be one of 'intercept', 'fixed', or 'random'.
            family (str): The name of the Family to retrieve. Must be a value
                defined internally. In the default config, this is one of
                'gaussian', 'bernoulli', 'poisson', or 't'.
        '''
        if dist is not None:
            if dist not in self.dists:
                raise ValueError(
                    "'%s' is not a valid distribution name." % dist)
            return self._get_prior(self.dists[dist])
        elif term is not None:
            if term not in self.terms:
                raise ValueError("'%s' is not a valid term type." % term)
            return self._get_prior(self.terms[term])
        elif family is not None:
            if family not in self.families:
                raise ValueError("'%s' is not a valid family name." % family)
            _f = self.families[family]
            prior = self._get_prior(_f['dist'])
            return Family(family, prior, _f['link'], _f['parent'])


class PriorScaler(object):

    # Default is 'wide'. The wide prior SD is sqrt(1/3) = .577 on the partial
    # corr scale, which is the SD of a flat prior over [-1,1].
    names = {
        'narrow': 0.2,
        'medium': 0.4,
        'wide': 3 ** -0.5,
        'superwide': 0.8
    }

    def __init__(self, model, taylor):
        self.model = model
        self.stats = model.dm_statistics if hasattr(model, 'dm_statistics') \
            else None
        self.dm = pd.DataFrame({lev: t.data[:, i]
                                for t in model.fixed_terms.values()
                                for i, lev in enumerate(t.levels)})
        self.priors = {}
        missing = 'drop' if self.model.dropna else 'none'
        self.mle = sm.GLM(endog=self.model.y.data, exog=self.dm,
                          family=self.model.family.smfamily(),
                          missing=missing).fit()
        self.taylor = taylor
        with open(join(dirname(__file__), 'config', 'derivs.txt'), 'r') as file:
            self.deriv = [next(file).strip('\n') for x in range(taylor+1)]

    def _get_slope_stats(self, exog, predictor, sd_corr, full_mod=None,
                         points=4):
        '''
        full_mod: statsmodels GLM to replace MLE model. For when 'predictor'
            is not in the fixed part of the model.
        points: number of points to use for LL approximation
        '''

        if full_mod is None:
            full_mod = self.mle

        # figure out which column of exog to drop for the null model
        keeps = [i for i, x in enumerate(list(exog.columns))
                 if not np.array_equal(predictor, exog[x].values.flatten())]
        i = [x for x in range(exog.shape[1]) if x not in keeps][0]

        # get log-likelihood values from beta=0 to beta=MLE
        values = np.linspace(0., full_mod.params[i], points)
        # if there are multiple predictors, use statsmodels to optimize the LL
        if keeps:
            null = [sm.GLM(endog=self.model.y.data, exog=exog,
                           family=self.model.family.smfamily()).fit_constrained(
                                str(exog.columns[i])+'='+str(val),
                                start_params=full_mod.params.values)
                    for val in values[:-1]]
            null = np.append(null, full_mod)
            ll = np.array([x.llf for x in null])
        # if just a single predictor, use statsmodels to evaluate the LL
        else:
            null = [self.model.family.smfamily().loglike(
                np.squeeze(self.model.y.data), val*predictor)
                for val in values[:-1]]
            ll = np.append(null, full_mod.llf)

        # compute params of quartic approximatino to log-likelihood
        # c: intercept, d: shift parameter
        # a: quartic coefficient, b: quadratic coefficient
        c, d = ll[-1], -np.asscalar(full_mod.params[i])
        X = np.matrix([(values+d)**4,
                       (values+d)**2]).T
        a, b = np.squeeze((np.linalg.inv(X.T * X) * X.T * (ll[:, None] - c)).A)

        # m, v: mean and variance of beta distribution of correlations
        # p, q: corresponding shape parameters of beta distribution
        m = .5
        v = sd_corr**2/4
        p = m*(m*(1-m)/v - 1)
        q = (1-m)*(m*(1-m)/v - 1)

        # function to return central moments of rescaled beta distribution
        def moment(k): return (2*p/(p+q))**k * hyp2f1(p, -k, p+q, (p+q)/p)

        # evaluate the derivatives of beta = f(correlation).
        # dict 'point' gives points about which to Taylor expand. We want to
        # expand about the mean (generally 0), but some of the derivatives
        # do not exist at 0. Evaluating at a point very close to 0 (e.g., .001)
        # generally gives good results, but the higher order the expansion, the
        # further from 0 we need to evaluate the derivatives, or they blow up.
        point = dict(zip(range(1, 14), 2**np.linspace(-1, 5, 13)/100))
        vals = dict(a=a, b=b, n=len(self.model.y.data), r=point[self.taylor])
        _deriv = [eval(x, globals(), vals) for x in self.deriv]

        # compute and return the approximate SD
        def term(i, j):
            return 1/np.math.factorial(i) * 1/np.math.factorial(j) \
                * _deriv[i] * _deriv[j] \
                * (moment(i+j) - moment(i)*moment(j))
        terms = [term(i, j) for i in range(1, self.taylor+1)
                 for j in range(1, self.taylor+1)]
        return np.array(terms).sum()**.5

    def _get_intercept_stats(self, add_slopes=True):
        # start with mean and variance of Y on the link scale
        mod = sm.GLM(endog=self.model.y.data,
                     exog=np.repeat(1, len(self.model.y.data)),
                     family=self.model.family.smfamily(),
                     missing='drop' if self.model.dropna else 'none').fit()
        mu = mod.params
        # multiply SE by sqrt(N) to turn it into (approx.) SD(Y) on link scale
        sd = (mod.cov_params()[0] * len(mod.mu))**.5

        # modify mu and sd based on means and SDs of slope priors.
        if len(self.model.fixed_terms) > 1 and add_slopes:
            means = np.array([x['mu'] for x in self.priors.values()])
            sds = np.array([x['sd'] for x in self.priors.values()])
            # add to intercept prior
            index = list(self.priors.keys())
            mu -= np.dot(means, self.stats['mean_x'][index])
            sd = (sd**2 + np.dot(sds**2, self.stats['mean_x'][index]**2))**.5

        return mu, sd

    def _scale_fixed(self, term):

        # these defaults are only defined for Normal priors
        if term.prior.name != 'Normal':
            return

        mu = []
        sd = []
        sd_corr = term.prior.scale
        for pred in term.data.T:
            mu += [0]
            sd += [self._get_slope_stats(exog=self.dm, predictor=pred,
                                         sd_corr=sd_corr)]

        # save and set prior
        for i, lev in enumerate(term.levels):
            self.priors.update({lev: {'mu': mu[i], 'sd': sd[i]}})
        term.prior.update(mu=np.array(mu), sd=np.array(sd))

    def _scale_intercept(self, term):

        # default priors are only defined for Normal priors
        if term.prior.name != 'Normal':
            return

        # get prior mean and SD for fixed intercept
        mu, sd = self._get_intercept_stats()

        # save and set prior
        term.prior.update(mu=mu, sd=sd)

    def _scale_random(self, term):

        # these default priors are only defined for HalfNormal priors
        if term.prior.args['sd'].name != 'HalfNormal':
            return

        sd_corr = term.prior.scale

        # recreate the corresponding fixed effect data
        fix_data = term.data.sum(axis=1)

        # handle intercepts and cell means
        if term.constant:
            mu, sd = self._get_intercept_stats()
            sd *= sd_corr
        # handle slopes
        else:
            exists = [x for x in self.dm.columns
                      if np.array_equal(fix_data, self.dm[x].values)]
            # handle case where there IS a corresponding fixed effect
            if exists and exists[0] in self.priors.keys():
                sd = self.priors[exists[0]]['sd']
            # handle case where there IS NOT a corresponding fixed effect
            else:
                # the usual case: add the random effect data as a fixed effect
                # in the design matrix
                if not exists:
                    fix_dataframe = pd.DataFrame(fix_data)
                    # things break if column names are integers (the default)
                    fix_dataframe.rename(
                        columns={c: '_'+str(c) for c in fix_dataframe.columns},
                        inplace=True)
                    exog = self.dm.join(fix_dataframe)
                # this handles the corner case where there technically is the
                # corresponding fixed effect, but the parameterization differs
                # between the fixed- and random-effect specification. usually
                # this means the fixed effects use cell-means coding but the
                # random effects use k-1 coding
                else:
                    group = term.name.split('|')[1]
                    exog = self.model.random_terms.values()
                    exog = [v.data.sum(1) for v in exog
                            if v.name.split('|')[-1] == group]
                    index = ['_'+str(i) for i in range(len(exog))]
                    exog = pd.DataFrame(exog, index=index).T
                # this will replace self.mle (which is missing predictors)
                missing = 'drop' if self.model.dropna else 'none'
                full_mod = sm.GLM(endog=self.model.y.data, exog=exog,
                                  family=self.model.family.smfamily(),
                                  missing=missing).fit()
                sd = self._get_slope_stats(exog=exog, predictor=fix_data,
                                           full_mod=full_mod, sd_corr=sd_corr)

        # set the prior SD.
        term.prior.args['sd'].update(sd=np.squeeze(np.atleast_1d(sd)))

    def scale(self):
        # classify all terms
        fixed_intercepts = [t for t in self.model.terms.values()
                            if not t.random and t.data.sum(1).var() == 0]
        fixed_slopes = [t for t in self.model.terms.values()
                        if not t.random and not t.data.sum(1).var() == 0]
        random_terms = [t for t in self.model.terms.values() if t.random]

        # arrange them in the order in which they should be initialized
        term_list = fixed_slopes + fixed_intercepts + random_terms
        term_types = ['fixed']*len(fixed_slopes) + \
            ['intercept']*len(fixed_intercepts) + \
            ['random']*len(random_terms)

        # initialize them in order
        for t, term_type in zip(term_list, term_types):
            if t.prior.scale is None:
                if not t.prior._auto_scale or not self.model.auto_scale:
                    continue
                t.prior.scale = 'wide'

            # Convert scale names to float
            if isinstance(t.prior.scale, string_types):
                t.prior.scale = PriorScaler.names[t.prior.scale]

            # scale it!
            getattr(self, '_scale_%s' % term_type)(t)



from abc import ABCMeta, abstractmethod


class BackEnd(object):

    '''
    Base class for BackEnd hierarchy.
    '''
    __metaclass__ = ABCMeta

    @abstractmethod
    def build(self):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def _convert_to_results(self):
        pass



from __future__ import absolute_import
from .base import BackEnd
from bambi.priors import Prior
from bambi.results import MCMCResults, PyMC3ADVIResults
from bambi.external.six import string_types
from collections import OrderedDict
import theano
import pymc3 as pm
import numpy as np
import re
from pymc3.model import TransformedRV


class PyMC3BackEnd(BackEnd):

    '''
    PyMC3 model-fitting back-end.
    '''

    # Available link functions
    links = {
        'identity': lambda x: x,
        'logit': theano.tensor.nnet.sigmoid,
        'inverse': theano.tensor.inv,
        'log': theano.tensor.log
    }

    dists = {
        'HalfFlat': pm.Bound(pm.Flat, lower=0)
    }

    def __init__(self):

        self.reset()

    def reset(self):
        '''
        Reset PyMC3 model and all tracked distributions and parameters.
        '''
        self.model = pm.Model()
        self.mu = None
        self.par_groups = {}

    def _build_dist(self, spec, label, dist, **kwargs):
        ''' Build and return a PyMC3 Distribution. '''
        if isinstance(dist, string_types):
            if hasattr(pm, dist):
                dist = getattr(pm, dist)
            elif dist in self.dists:
                dist = self.dists[dist]
            else:
                raise ValueError("The Distribution class '%s' was not "
                                 "found in PyMC3 or the PyMC3BackEnd." % dist)

        # Inspect all args in case we have hyperparameters
        def _expand_args(k, v, label):
            if isinstance(v, Prior):
                label = '%s_%s' % (label, k)
                return self._build_dist(spec, label, v.name, **v.args)
            return v

        kwargs = {k: _expand_args(k, v, label) for (k, v) in kwargs.items()}

        # Non-centered parameterization for hyperpriors
        if spec.noncentered and 'sd' in kwargs and 'observed' not in kwargs \
                and isinstance(kwargs['sd'], pm.model.TransformedRV):
            old_sd = kwargs['sd']
            _offset = pm.Normal(label + '_offset', mu=0, sd=1,
                                shape=kwargs['shape'])
            return pm.Deterministic(label, _offset * old_sd)

        return dist(label, **kwargs)

    def build(self, spec, reset=True):
        '''
        Compile the PyMC3 model from an abstract model specification.
        Args:
            spec (Model): A bambi Model instance containing the abstract
                specification of the model to compile.
            reset (bool): if True (default), resets the PyMC3BackEnd instance
                before compiling.
        '''
        if reset:
            self.reset()

        with self.model:

            self.mu = 0.

            for t in spec.terms.values():

                data = t.data
                label = t.name
                dist_name = t.prior.name
                dist_args = t.prior.args

                n_cols = t.data.shape[1]

                coef = self._build_dist(spec, label, dist_name,
                                        shape=n_cols, **dist_args)

                if t.random:
                    self.mu += coef[t.group_index][:, None] * t.predictor
                else:
                    self.mu += pm.math.dot(data, coef)[:, None]

            y = spec.y.data
            y_prior = spec.family.prior
            link_f = spec.family.link
            if not callable(link_f):
                link_f = self.links[link_f]
            y_prior.args[spec.family.parent] = link_f(self.mu)
            y_prior.args['observed'] = y
            y_like = self._build_dist(spec, spec.y.name, y_prior.name,
                                      **y_prior.args)

            self.spec = spec

    def _get_transformed_vars(self):
        rvs = self.model.unobserved_RVs
        # identify the variables that pymc3 back-transformed to original scale
        trans = [var.name for var in rvs if isinstance(var, TransformedRV)]
        # find the corresponding transformed variables
        trans = set(x.name for x in rvs if any([t in x.name for t in trans])) \
            - set(trans)
        # add any "centered" random effects to the list
        for varname in [var.name for var in rvs]:
            if re.search(r'_offset$', varname) is not None: trans.add(varname)

        return trans

    def run(self, start=None, method='mcmc', init='advi', n_init=50000,
            find_map=False, **kwargs):
        '''
        Run the PyMC3 MCMC sampler.
        Args:
            start: Starting parameter values to pass to sampler; see
                pm.sample() documentation for details.
            method: The method to use for fitting the model. By default,
                'mcmc', in which case the PyMC3 sampler will be used.
                Alternatively, 'advi', in which case the model will be fitted
                using  automatic differentiation variational inference as
                implemented in PyMC3.
            init: Initialization method (see PyMC3 sampler documentation).
                In PyMC3, this defaults to 'advi', but we set it to None.
            n_init: Number of initialization iterations if init = 'advi' or
                'nuts'. Default is kind of in PyMC3 for the kinds of models
                we expect to see run with bambi, so we lower it considerably.
            find_map (bool): whether or not to use the maximum a posteriori
                estimate as a starting point; passed directly to PyMC3.
            kwargs (dict): Optional keyword arguments passed onto the sampler.
        Returns: A PyMC3ModelResults instance.
        '''

        if method == 'mcmc':
            samples = kwargs.pop('samples', 1000)
            njobs = kwargs.pop('chains', 1)
            with self.model:
                if start is None and find_map:
                    start = pm.find_MAP()
                self.trace = pm.sample(samples, start=start, init=init,
                                       n_init=n_init, njobs=njobs, **kwargs)
            return self._convert_to_results()

        elif method == 'advi':
            with self.model:
                self.advi_params = pm.variational.advi(start, **kwargs)
            return PyMC3ADVIResults(self.spec, self.advi_params,
                                    transformed_vars=self._get_transformed_vars())

    def _convert_to_results(self):
        # grab samples as big, unlabelled array
        # dimensions 0, 1, 2 = samples, chains, variables
        data = np.array([np.array([np.atleast_2d(x.T).T[:, i]
                         for x in self.trace._straces[j].samples.values()
                         for i in range(np.atleast_2d(x.T).T.shape[1])])
                         for j in range(len(self.trace._straces))])
        data = np.swapaxes(np.swapaxes(data, 0, 1), 0, 2)

        # arrange var_shapes dictionary in same order as samples dictionary
        shapes = OrderedDict()
        for key in self.trace._straces[0].samples:
            shapes[key] = self.trace._straces[0].var_shapes[key]

        # grab info necessary for making samplearray pretty
        names = list(shapes.keys())
        dims = list(shapes.values())
        def get_levels(key, value):
            if len(value):
                # fixed effects
                if not self.spec.terms[re.sub(r'_offset$', '', key)].random:
                    return self.spec.terms[key].levels
                # random effects
                else:
                    re1 = re.match(r'(.+)(?=_offset)(_offset)', key)
                    # handle "centered" terms
                    if re1 is None:
                        return [key.split('|')[0]+'|'+x
                                for x in self.spec.terms[key].levels]
                    # handle "non-centered" terms
                    else:
                        return ['{}|{}_offset{}'.format(key.split('|')[0],
                                *re.match(r'^(.+)(\[.+\])$', x).groups())
                                for x in self.spec.terms[re1.group(1)].levels]
            else:
                return [key]
        levels = sum([get_levels(k, v) for k, v in shapes.items()], [])

        # instantiate
        return MCMCResults(model=self.spec, data=data, names=names, dims=dims,
                           levels=levels,
                           transformed_vars=self._get_transformed_vars())



from abc import abstractmethod, ABCMeta
from bambi.external.six import string_types
from bambi import diagnostics as bmd
import re
import warnings
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from .utils import listify
try:
    import pymc3 as pm
    if hasattr(pm.plots, 'kdeplot_op'):
        pma = pm.plots
    else:
        pma = pm.plots.artists
except:
    pma = None


__all__ = ['MCMCResults', 'PyMC3ADVIResults']


class ModelResults(object):

    '''
    Base class for ModelResults hierarchy.
    Args:
        model (Model): a bambi Model instance specifying the model.
    '''

    __metaclass__ = ABCMeta

    def __init__(self, model):

        self.model = model
        self.terms = list(model.terms.values())
        self.diagnostics = model._diagnostics \
            if hasattr(model, '_diagnostics') else None

    @abstractmethod
    def plot(self):
        pass

    @abstractmethod
    def summary(self):
        pass


class MCMCResults(ModelResults):

    '''
    Holds sampler results; provides slicing, plotting, and summarization tools.
    Args:
        model (Model): a bambi Model instance specifying the model.
        data (numpy array): Raw storage of MCMC samples in array with
            dimensions 0, 1, 2 = samples, chains, variables
        names (list): Names of all Terms.
        dims (list): Numbers of levels for all Terms.
        levels (list): Names of all levels for all Terms.
        transformed (list): Optional list of variable names to treat as
            transformed--and hence, to exclude from the output by default.
    '''

    def __init__(self, model, data, names, dims, levels,
                 transformed_vars=None):
        # store the arguments
        self.data = data
        self.names = names
        self.dims = dims
        self.levels = levels
        self.transformed_vars = transformed_vars

        # compute basic stuff to use later
        self.n_samples = data.shape[0]
        self.n_chains = data.shape[1]
        self.n_params = data.shape[2]
        self.n_terms = len(names)
        if transformed_vars is not None:
            utv = list(set(names) - set(transformed_vars))
        else:
            utv = names
        self.untransformed_vars = utv
        # this keeps track of which columns in 'data' go with which terms
        self.index = np.cumsum(
            [0] + [x[0] if len(x) else 1 for x in dims][:-1])

        # build level_dict: dictionary of lists containing levels of each Term
        level_dict = {}
        for i, name, dim in zip(self.index, names, dims):
            dim = dim[0] if len(dim) else 1
            level_dict[name] = levels[i:(i+dim)]
        self.level_dict = level_dict

        super(MCMCResults, self).__init__(model)

    def __getitem__(self, idx):
        '''
        If a variable name, return MCMCResults with only that variable
            e.g., fit['subject']
        If a list of variable names, return MCMCResults with those variables
            e.g., fit[['subject','item]]
        If a slice, return MCMCResults with sliced samples
            e.g., fit[500:]
        If a tuple, return MCMCResults with those variables sliced
            e.g., fit[['subject','item'], 500:] OR fit[500:, ['subject','item']]
        '''

        if isinstance(idx, slice):
            var = self.names
            vslice = idx
        elif isinstance(idx, string_types):
            var = [idx]
            vslice = slice(0, self.n_samples)
        elif isinstance(idx, list):
            if not all([isinstance(x, string_types) for x in idx]):
                raise ValueError("If passing a list, all elements must be "
                                 "parameter names.")
            var = idx
            vslice = slice(0, self.n_samples)
        elif isinstance(idx, tuple):
            if len(idx) > 2:
                raise ValueError("Only two arguments can be passed. If you "
                                 "want to select multiple parameters and a "
                                 "subset of samples, pass a slice and a list "
                                 "of parameter names.")
            vslice = [i for i, x in enumerate(idx) if isinstance(x, slice)]
            if not len(vslice):
                raise ValueError("At least one argument must be a slice. If "
                                 "you want to select multiple parameters by "
                                 "name, pass a list (not a tuple) of names.")
            if len(vslice) > 1:
                raise ValueError("Slices can only be applied "
                                 "over the samples dimension.")
            var = idx[1 - vslice[0]]
            vslice = idx[vslice[0]]
            if not isinstance(var, (list, tuple)):
                var = [var]
        else:
            raise ValueError("Unrecognized index type.")

        # do slicing/selection and return subsetted copy of MCMCResults
        levels = sum([self.level_dict[v] for v in var], [])
        level_iloc = [self.levels.index(x) for x in levels]
        var_iloc = [self.names.index(v) for v in var]
        return MCMCResults(model=self.model,
                           data=self.data[vslice, :, level_iloc], names=var,
                           dims=[self.dims[x] for x in var_iloc],
                           levels=levels,
                           transformed_vars=self.transformed_vars)

    def get_chains(self, indices):
        # Return copy of self but only for chains with the passed indices
        if not isinstance(indices, (list, tuple)):
            indices = [indices]
        return MCMCResults(model=self.model, data=self.data[:, indices, :],
                           names=self.names, dims=self.dims,
                           levels=self.levels,
                           transformed_vars=self.transformed_vars)

    def _filter_names(self, varnames=None, ranefs=False, transformed=False):
        names = self.untransformed_vars if not transformed else self.names
        if varnames is not None:
            names = [n for n in names if n in listify(varnames)]
        if not ranefs:
            names = [x for x in names if re.sub(r'_offset$', '', x)
                     not in self.model.random_terms]
        return names

    def plot(self, varnames=None, ranefs=True, transformed=False,
             combined=False, hist=False, bins=20, kind='trace'):
        '''
        Plots posterior distributions and sample traces. Basically a wrapper
        for pm.traceplot() plus some niceties, based partly on code from:
        https://pymc-devs.github.io/pymc3/notebooks/GLM-model-selection.html
        Args:
            varnames (list): List of variable names to plot. If None, all
                eligible variables are plotted.
            ranefs (bool): If True (default), shows trace plots for
                individual random effects.
            transformed (bool): If False (default), excludes internally
                transformed variables from plotting.
            combined (bool): If True, concatenates all chains into one before
                plotting. If False (default), plots separately lines for
                each chain (on the same axes).
            hist (bool): If True, plots a histogram for each fixed effect,
                in addition to the kde plot. To prevent visual clutter,
                histograms are never plotted for random effects.
            bins (int): If hist is True, the number of bins in the histogram.
                Ignored if hist is False.
            kind (str): Either 'trace' (default) or 'priors'. If 'priors',
                this just internally calls Model.plot()
        '''

        def _plot_row(data, row, title, hist=True):
            # density plot
            axes[row, 0].set_title(title)
            if pma is not None:
                arr = np.atleast_2d(data.values.T).T
                pma.kdeplot_op(axes[row, 0], arr)
            else:
                data.plot(kind='kde', ax=axes[row, 0], legend=False)

            # trace plot
            axes[row, 1].set_title(title)
            data.plot(kind='line', ax=axes[row, 1], legend=False)
            # histogram
            if hist:
                if pma is not None:
                    pma.histplot_op(axes[row, 0], arr)
                else:
                    data.plot(kind='hist', ax=axes[row, 0], legend=False,
                              normed=True, bins=bins)

        if kind == 'priors':
            return self.model.plot(varnames)

        # count the total number of rows in the plot
        names = self._filter_names(varnames, ranefs, transformed)
        random = [re.sub(r'_offset$', '', x) in self.model.random_terms
                  for x in names]
        rows = sum([len(self.level_dict[p]) if not r else 1
                    for p, r in zip(names, random)])

        # make the plot!
        fig, axes = plt.subplots(rows, 2, figsize=(12, 2*rows))
        if rows == 1:
            axes = np.array([axes])  # For consistent 2D indexing

        _select_args = {'ranefs': ranefs, 'transformed': transformed}

        # Create list of chains (for combined, just one list w/ all chains)
        chains = list(range(self.n_chains))
        if combined:
            chains = [chains]

        for c in chains:

            row = 0

            for p in names:

                df = self[p].to_df(chains=c, **_select_args)
                # if p == 'floor|county_sd':

                # fixed effects
                if re.sub(r'_offset$', '', p) not in self.model.random_terms:
                    for lev in self.level_dict[p]:
                        lev_df = df[lev]
                        _plot_row(lev_df, row, lev, hist)
                        row += 1

                # random effects
                else:
                    _plot_row(df, row, p, hist=False)  # too much clutter
                    row += 1

        fig.tight_layout()

        # For bernoulli models, tell user which event is being modeled
        if self.model.family.name == 'bernoulli':
            event = next(i for i, x in enumerate(self.model.y.data.flatten())
                         if x > .99)
            warnings.warn('Modeling the probability that {}==\'{}\''.format(
                self.model.y.name,
                str(self.model.data[self.model.y.name][event])))

        return axes

    def _hpd_interval(self, x, width):
        """
        Code adapted from pymc3.stats.calc_min_interval:
        https://github.com/pymc-devs/pymc3/blob/master/pymc3/stats.py
        """
        x = np.sort(x)
        n = len(x)

        interval_idx_inc = int(np.floor(width * n))
        n_intervals = n - interval_idx_inc
        interval_width = x[interval_idx_inc:] - x[:n_intervals]

        if len(interval_width) == 0:
            raise ValueError('Too few elements for interval calculation')

        min_idx = np.argmin(interval_width)
        hdi_min = x[min_idx]
        hdi_max = x[min_idx + interval_idx_inc]

        index = ['hpd{}_{}'.format(width, x) for x in ['lower', 'upper']]
        return pd.Series([hdi_min, hdi_max], index=index)

    def summary(self, varnames=None, ranefs=False, transformed=False, hpd=.95,
                quantiles=None, diagnostics=['effective_n', 'gelman_rubin']):
        '''
        Returns a DataFrame of summary/diagnostic statistics for the
        parameters.
        Args:
            varnames (list): List of variable names to include; if None
                (default), all eligible variables are included.
            ranefs (bool): Whether or not to include random effects in the
                summary. Default is False.
            transformed (bool): Whether or not to include internally
                transformed variables in the summary. Default is False.
            hpd (float, between 0 and 1): Show Highest Posterior Density (HPD)
                intervals with specified width/proportion for all parameters.
                If None, HPD intervals are suppressed.
            quantiles (float [or list of floats] between 0 and 1): Show
                specified quantiles of the marginal posterior distributions for
                all parameters. If None (default), no quantiles are shown.
            diagnostics (list): List of functions to use to compute convergence
                diagnostics for all parameters. Each element can be either a
                callable or a string giving the name of a function in the
                diagnostics module. Valid strings are 'gelman_rubin' and
                'effective_n'. Functions must accept a MCMCResults object as
                the sole input, and return a DataFrame with one labeled row per
                parameter. If None, no convergence diagnostics are computed.
        '''
        samples = self.to_df(varnames, ranefs, transformed)

        # build the basic DataFrame
        df = pd.DataFrame({'mean': samples.mean(0), 'sd': samples.std(0)})

        # add user-specified quantiles
        if quantiles is not None:
            if not isinstance(quantiles, (list, tuple)):
                quantiles = [quantiles]
            qnames = ['q' + str(q) for q in quantiles]
            df = df.merge(samples.quantile(quantiles).set_index([qnames]).T,
                          left_index=True, right_index=True)

        # add HPD intervals
        if hpd is not None:
            df = df.merge(samples.apply(self._hpd_interval, axis=0,
                          width=hpd).T, left_index=True, right_index=True)

        # add convergence diagnostics
        if diagnostics is not None:
            _names = self._filter_names(ranefs=ranefs, transformed=transformed)
            _self = self[_names]
            if self.n_chains > 1:
                for diag in diagnostics:
                    if isinstance(diag, string_types):
                        diag = getattr(bmd, diag)
                    df = df.merge(diag(_self), left_index=True,
                                  right_index=True)
            else:
                warnings.warn('Multiple MCMC chains are required in order '
                              'to compute convergence diagnostics.')

        # For bernoulli models, tell user which event is being modeled
        if self.model.family.name == 'bernoulli':
            event = next(i for i, x in enumerate(self.model.y.data.flatten())
                         if x > .99)
            warnings.warn('Modeling the probability that {}==\'{}\''.format(
                self.model.y.name,
                str(self.model.data[self.model.y.name][event])))

        return df

    def to_df(self, varnames=None, ranefs=False, transformed=False,
              chains=None):
        '''
        Returns the MCMC samples in a nice, neat pandas DataFrame with all
        MCMC chains concatenated.
        Args:
            varnames (list): List of variable names to include; if None
                (default), all eligible variables are included.
            ranefs (bool): Whether or not to include random effects in the
                returned DataFrame. Default is True.
            transformed (bool): Whether or not to include internally
                transformed variables in the result. Default is False.
            chains (int, list): Index, or list of indexes, of chains to
                concatenate. E.g., [1, 3] would concatenate the first and
                third chains, and ignore any others. If None (default),
                concatenates all available chains.
        '''
        # filter out unwanted variables
        names = self._filter_names(varnames, ranefs, transformed)

        # concatenate the (pre-sliced) chains
        if chains is None:
            chains = list(range(self.n_chains))
        chains = listify(chains)
        data = [self.data[:, i, :] for i in chains]
        data = np.concatenate(data, axis=0)

        # construct the trace DataFrame
        df = sum([self.level_dict[x] for x in names], [])
        df = pd.DataFrame({x: data[:, self.levels.index(x)] for x in df})

        return df


class PyMC3ADVIResults(ModelResults):

    '''
    Holds PyMC3 ADVI results and provides plotting and summarization tools.
    Args:
        model (Model): a bambi Model instance specifying the model.
        params (MultiTrace): ADVI parameters returned by PyMC3.
    '''

    def __init__(self, model, params):

        self.means = params['means']
        self.sds = params['stds']
        self.elbo_vals = params['elbo_vals']
        super(PyMC3ADVIResults, self).__init__(model)



from __future__ import absolute_import
from .base import BackEnd
from bambi.priors import Prior
from bambi.results import MCMCResults
import numpy as np
import pymc3 as pm
import re
try:
    import pystan as ps
except:
    ps = None
from six import string_types


class StanBackEnd(BackEnd):

    '''
    Stan/PyStan model-fitting back-end.
    '''

    # distribution names and arg names should match those in priors.json,
    dists = {
        'Normal': {'name': 'normal', 'args': ['#mu', '#sd']},
        'Bernoulli': {'name': 'bernoulli', 'args': ['#p']},
        'Poisson': {'name': 'poisson', 'args': ['#mu']},
        'Cauchy': {'name': 'cauchy', 'args': ['#alpha', '#beta']},
        'HalfNormal': {'name': 'normal', 'args': ['0', '#sd'],
                       'bounds': '<lower=0>'},
        'HalfCauchy': {'name': 'cauchy', 'args': ['0', '#beta'],
                       'bounds': '<lower=0>'},
        # for Uniform, the bounds are the parameters. _map_dist fills these in
        'Uniform': {'name': 'uniform', 'args': ['#lower', '#upper'],
                    'bounds': '<lower={}, upper={}>'},
        'Flat': {'name': None, 'args': []},
        'HalfFlat':  {'name': None, 'args': [], 'bounds': '<lower=0>'}
    }

    # maps from spec.family.name to 'dists' dictionary
    # and gives the Stan variable type for the response
    families = {
        'gaussian': {'name': 'Normal', 'type': float,
            'format': ['vector[N]', '']},
        'bernoulli': {'name': 'Bernoulli', 'type': int,
            'format': ['int', '[N]']},
        'poisson': {'name': 'Poisson', 'type': int,
            'format': ['int', '[N]']}
    }

    # maps from spec.family.link to Stan inverse link function name
    links = {
        'identity': '',
        'logit': 'inv_logit',
        'log': 'exp'
    }

    def __init__(self):
        if ps is None:
            raise ImportError("Could not import PyStan; please make sure it's "
                              "installed.")
        self.reset()

    def reset(self):
        '''
        Reset Stan model and all tracked distributions and parameters.
        '''
        self.parameters = []
        self.transformed_parameters = []
        self.expressions = []
        self.data = []
        self.transformed_data = []
        self.X = {}
        self.model = []
        self.mu_cont = []
        self.mu_cat = []
        self._original_names = {}
        # variables to suppress in output. Stan uses limited set for variable
        # names, so track variable names we may need to simplify for the model
        # code and then sub back later.
        self._suppress_vars = ['yhat', 'lp__']

    def build(self, spec, reset=True):
        '''
        Compile the Stan model from an abstract model specification.
        Args:
            spec (Model): A bambi Model instance containing the abstract
                specification of the model to compile.
            reset (bool): if True (default), resets the StanBackEnd instance
                before compiling.
        '''
        if reset:
            self.reset()

        n_cases = len(spec.y.data)
        self.data.append('int<lower=1> N;')
        self.X['N'] = n_cases

        def _sanitize_name(name):
            ''' Stan only allows alphanumeric chars and underscore, and
            variable names must begin with a letter. Additionally, Stan
            reserves a few hundred strings that can't be used as variable
            names. So to play it safe, we replace all invalid chars with '_',
            and prepend all variables with 'b_'. We substitute the original
            names back in later. '''
            if name in self._original_names:
                return name
            clean = 'b_' + re.sub('[^a-zA-Z0-9\_]+', '_', name)
            self._original_names[clean] = name
            return clean

        def _map_dist(dist, **kwargs):
            ''' Maps PyMC3 distribution names and attrs in the Prior object
            to the corresponding Stan names and argument order. '''
            if dist not in self.dists:
                raise ValueError("There is no distribution named '%s' "
                                 "in Stan." % dist)

            stan_dist = self.dists[dist]
            dist_name = stan_dist['name']
            dist_args = stan_dist['args']
            dist_bounds = stan_dist.get('bounds', '')

            # Flat/HalfFlat/undefined priors are handled separately
            if dist_name is None:
                return None, dist_bounds

            lookup_args = [a[1:] for a in dist_args if a.startswith('#')]
            missing = set(lookup_args) - set(list(kwargs.keys()))
            if missing:
                raise ValueError("The following mandatory parameters of "
                                 "the %s distribution are missing: %s."
                                 % (dist, missing))

            # Named arguments to take from the Prior object are denoted with
            # a '#'; otherwise we take the value in the self.dists dict as-is.
            dp = [kwargs[p[1:]] if p.startswith('#') else p for p in dist_args]

            # Sometimes we get numpy arrays at this stage, so convert to float
            dp = [float(p.ravel()[0]) if isinstance(p, np.ndarray) else p
                  for p in dp]

            dist_term = '%s(%s)' % (dist_name, ', '.join([str(p) for p in dp]))

            # handle Uniform variables, for which the bounds are the parameters
            if dist_name=='uniform':
                dist_bounds = dist_bounds.format(*dp)

            return dist_term, dist_bounds

        def _add_data(name, data, term):
            ''' Add all model components that directly touch or relate to data.
            '''
            if data.shape[1] == 1:
                # For random effects, index into grouping variable
                if n_cols > 1:
                    index_name = _sanitize_name('%s_grp_ind' % name)
                    self.data.append('int %s[N];' % index_name)
                    self.X[index_name] = t.group_index + 1  # 1-based indexing
                predictor = 'vector[N] %s;'
            else:
                predictor = ('matrix[N, %d]' % (n_cols)) + ' %s;'

            data_name = _sanitize_name('%s_data' % name)
            var_name = _sanitize_name(name)
            self.data.append(predictor % data_name)
            self.X[data_name] = data.squeeze()

            if data.shape[1] == 1 and n_cols > 1:
                code = '%s[%s[n]] * %s[n]' % (var_name, index_name, data_name)
                self.mu_cat.append(code)
            else:
                self.mu_cont.append('%s * %s' % (data_name, var_name))

        def _add_parameters(name, dist_name, n_cols, **dist_args):
            ''' Add all model components related to latent parameters. We
            handle these separately from the data components, as the parameters
            can have nested specifications (in the case of random effects). '''

            def _expand_args(k, v, name):
                if isinstance(v, Prior):
                    name = _sanitize_name('%s_%s' % (name, k))
                    return _add_parameters(name, v.name, 1, **v.args)
                return v

            kwargs = {k: _expand_args(k, v, name)
                      for (k, v) in dist_args.items()}

            _dist, _bounds = _map_dist(dist_name, **kwargs)

            if n_cols == 1:
                stan_par = 'real'
            else:
                stan_par = 'vector[%d]' % n_cols

            var_name = _sanitize_name(name)

            # self.parameters.append('%s%s %s;' % (stan_par, _bounds, var_name))

            # non-centered parameterization
            if spec.noncentered and 'sd' in kwargs and \
                    isinstance(kwargs['sd'], string_types):
                offset_name = _sanitize_name(name + '_offset')
                offset = 'vector[%d] %s;' % (n_cols, offset_name)
                self.parameters.append(offset)
                self.model.append('%s ~ normal(0, 1);' % offset_name)
                self.transformed_parameters.append('%s%s %s;' % (stan_par,
                                                                 _bounds,
                                                                 var_name))
                trans = '%s = multiply(%s, %s);' % (var_name, offset_name,
                                                    kwargs['sd'])
                self.expressions.append(trans)

            else:
                self.parameters.append('%s%s %s;' % (stan_par, _bounds,
                                                     var_name))
                if _dist is not None:
                    self.model.append('%s ~ %s;' % (var_name, _dist))

            return name

        for t in spec.terms.values():

            data = t.data
            label = t.name
            dist_name = t.prior.name
            dist_args = t.prior.args

            n_cols = data.shape[1]

            if t.random:
                data = t.predictor

            # Add to Stan model
            _add_data(label, data, t)
            _add_parameters(label, dist_name, n_cols, **dist_args)

        # yhat
        self.transformed_parameters.append('vector[N] yhat;')
        if self.mu_cont:
            yhat_cont = 'yhat = %s;' % ' + '.join(self.mu_cont)
            self.expressions.append(yhat_cont)
        else:
            self.mu_cat.insert(0, '0')

        if self.mu_cat:
            loops = ('for (n in 1:N)\n\t\tyhat[n] = yhat[n] + %s'
                     % ' + '.join(self.mu_cat) + ';\n\t')
            self.expressions.append(loops)

        # Add expressions that go in transformed parameter block (they have
        # to come after variable definitions)
        self.transformed_parameters += self.expressions

        # add response variable (y)
        _response_format = self.families[spec.family.name]['format']
        self.data.append('{} y{};'.format(*_response_format))

        # add response distribution parameters other than the location
        # parameter
        for k, v in spec.family.prior.args.items():
            if k != spec.family.parent and isinstance(v, Prior):
                _bounds = _map_dist(v.name, **v.args)[1]
                _param = 'real{} {}_{};'.format(_bounds, spec.y.name, k)
                self.parameters.append(_param)

        # specify the response distribution
        _response_dist = self.families[spec.family.name]['name']
        _response_args = '{}(yhat)'.format(self.links[spec.family.link])
        _response_args = {spec.family.parent: _response_args}
        for k, v in spec.family.prior.args.items():
            if k != spec.family.parent:
                _response_args[k] = '{}_{}'.format(spec.y.name, k) \
                    if isinstance(v, Prior) else str(v)
        _dist = _map_dist(_response_dist, **_response_args)[0]
        self.model.append('y ~ {};'.format(_dist))

        # add the data
        _response_type = self.families[spec.family.name]['type']
        self.X['y'] = spec.y.data.astype(_response_type).squeeze()

        # Construct the stan script
        def format_block(name):
            key = name.replace(' ', '_')
            els = ''.join(['\t%s\n' % e for e in getattr(self, key)])
            return '%s {\n%s}\n' % (name, els)

        blocks = ['data', 'transformed data', 'parameters',
                  'transformed parameters', 'model']
        self.model_code = ''.join([format_block(bl) for bl in blocks])
        self.spec = spec
        self.stan_model = ps.StanModel(model_code=self.model_code)

    def run(self, samples=1000, chains=1, **kwargs):
        '''
        Run the Stan sampler.
        Args:
            samples (int): Number of samples to obtain (in each chain).
            chains (int): Number of chains to use.
            kwargs (dict): Optional keyword arguments passed onto the PyStan
                StanModel.sampling() call.
        Returns: A PyMC3ModelResults instance.
        '''
        self.fit = self.stan_model.sampling(data=self.X, iter=samples,
                                            chains=chains, **kwargs)
        return self._convert_to_results()

    def _convert_to_results(self):
        # grab samples as big, unlabelled array
        # dimensions 0, 1, 2 = samples, chains, variables
        data = self.fit.extract(permuted=False, inc_warmup=True)     

        # grab info necessary for making samplearray pretty
        names = [self._original_names[x] \
                 if x in self._original_names else x \
                 for x in self.fit.sim['pars_oi']]
        dims = [tuple(x) for x in self.fit.sim['dims_oi']]

        def replace_name(name):
            tname = re.sub(r'\[.+\]$', '', name)
            if tname in self._original_names:
                lev = re.search(r'\[(.+)\]$', name)
                if lev is not None:
                    tname = self._original_names[tname]
                    # detect and handle offset terms
                    rgx = re.search(r'_offset$', tname)
                    if rgx is not None:
                        tname = tname[:rgx.start()]
                    lev = self.spec.terms[tname].levels[int(lev.group(1))]
                    split = tname.split('|')
                    # random effects
                    if len(split) == 2:
                        if rgx is None:
                            return '{}|{}'.format(split[0], lev)
                        else:
                            lev = re.search(r'(\[.+\])', lev).group(0)
                            return '{}|{}_offset{}'.format(*(split + [lev]))
                    # fixed effects
                    elif len(split) == 1:
                        return lev
                else:
                    return self._original_names[name]
            else:
                return name
        levels = [replace_name(x) for x in self.fit.sim['fnames_oi']]
        # Suppress non-centered parameterization parameters
        self._suppress_vars.extend([n for n in names if '_offset' in n])

        # instantiate
        return MCMCResults(model=self.spec, data=data, names=names, dims=dims,
                           levels=levels, transformed_vars=self._suppress_vars)
